@@ -17,6 +17,9 @@ const ExtensionUtils = imports.misc.extensionUtils;
 const Me = ExtensionUtils.getCurrentExtension();
 const Convenience = Me.imports.convenience;
 
+const Gettext = imports.gettext.domain('gnome-shell-extensions');
+const _ = Gettext.gettext;
+
 const ICON_TEXTURE_SIZE = 24;
 const DND_ACTIVATE_TIMEOUT = 500;
 
@@ -44,6 +47,16 @@ function _openMenu(menu) {
         menu.actor.navigate_focus(null, Gtk.DirectionType.TAB_FORWARD, false);
 }
 
+function _onMenuStateChanged(menu, isOpen) {
+    if (isOpen)
+        return;
+
+    let [x, y,] = global.get_pointer();
+    let actor = global.stage.get_actor_at_pos(Clutter.PickMode.REACTIVE, x, y);
+    if (windowList.actor.contains(actor))
+        actor.sync_hover();
+}
+
 
 const WindowContextMenu = new Lang.Class({
     Name: 'WindowContextMenu',
@@ -184,6 +197,7 @@ const WindowButton = new Lang.Class({
 
         this._menuManager = new PopupMenu.PopupMenuManager(this);
         this._contextMenu = new WindowContextMenu(this.actor, this.metaWindow);
+        this._contextMenu.connect('open-state-changed', _onMenuStateChanged);
         this._contextMenu.actor.hide();
         this._menuManager.addMenu(this._contextMenu);
         Main.uiGroup.add_actor(this._contextMenu.actor);
@@ -205,6 +219,17 @@ const WindowButton = new Lang.Class({
         this._updateStyle();
     },
 
+    get active() {
+        return this.actor.has_style_class_name('focused');
+    },
+
+    activate: function() {
+        if (this.active)
+            return;
+
+        this._onClicked(this.actor, 1);
+    },
+
     _onClicked: function(actor, button) {
         if (this._contextMenu.isOpen) {
             this._contextMenu.close();
@@ -252,7 +277,7 @@ const WindowButton = new Lang.Class({
     _onDestroy: function() {
         global.window_manager.disconnect(this._switchWorkspaceId);
         global.display.disconnect(this._notifyFocusId);
-        this._contextMenu.actor.destroy();
+        this._contextMenu.destroy();
     }
 });
 
@@ -268,7 +293,7 @@ const AppContextMenu = new Lang.Class({
 
         this._minimizeItem = new PopupMenu.PopupMenuItem(_("Minimize all"));
         this._minimizeItem.connect('activate', Lang.bind(this, function() {
-            this._app.get_windows().forEach(function(w) {
+            this._getWindowList().forEach(function(w) {
                 w.minimize();
             });
         }));
@@ -276,7 +301,7 @@ const AppContextMenu = new Lang.Class({
 
         this._unminimizeItem = new PopupMenu.PopupMenuItem(_("Unminimize all"));
         this._unminimizeItem.connect('activate', Lang.bind(this, function() {
-            this._app.get_windows().forEach(function(w) {
+            this._getWindowList().forEach(function(w) {
                 w.unminimize();
             });
         }));
@@ -284,7 +309,7 @@ const AppContextMenu = new Lang.Class({
 
         this._maximizeItem = new PopupMenu.PopupMenuItem(_("Maximize all"));
         this._maximizeItem.connect('activate', Lang.bind(this, function() {
-            this._app.get_windows().forEach(function(w) {
+            this._getWindowList().forEach(function(w) {
                 w.maximize(Meta.MaximizeFlags.HORIZONTAL |
                            Meta.MaximizeFlags.VERTICAL);
             });
@@ -293,7 +318,7 @@ const AppContextMenu = new Lang.Class({
 
         this._unmaximizeItem = new PopupMenu.PopupMenuItem(_("Unmaximize all"));
         this._unmaximizeItem.connect('activate', Lang.bind(this, function() {
-            this._app.get_windows().forEach(function(w) {
+            this._getWindowList().forEach(function(w) {
                 w.unmaximize(Meta.MaximizeFlags.HORIZONTAL |
                              Meta.MaximizeFlags.VERTICAL);
             });
@@ -302,15 +327,22 @@ const AppContextMenu = new Lang.Class({
 
         let item = new PopupMenu.PopupMenuItem(_("Close all"));
         item.connect('activate', Lang.bind(this, function() {
-            this._app.get_windows().forEach(function(w) {
+            this._getWindowList().forEach(function(w) {
                 w.delete(global.get_current_time());
             });
         }));
         this.addMenuItem(item);
     },
 
+    _getWindowList: function() {
+        let workspace = global.screen.get_active_workspace();
+        return this._app.get_windows().filter(function(win) {
+            return !win.skip_taskbar && win.located_on_workspace(workspace);
+        });
+    },
+
     open: function(animate) {
-        let windows = this._app.get_windows();
+        let windows = this._getWindowList();
         this._minimizeItem.actor.visible = windows.some(function(w) {
             return !w.minimized;
         });
@@ -360,6 +392,7 @@ const AppButton = new Lang.Class({
 
         this._menuManager = new PopupMenu.PopupMenuManager(this);
         this._menu = new PopupMenu.PopupMenu(this.actor, 0.5, St.Side.BOTTOM);
+        this._menu.connect('open-state-changed', _onMenuStateChanged);
         this._menu.actor.hide();
         this._menu.connect('activate', Lang.bind(this, this._onMenuActivate));
         this._menuManager.addMenu(this._menu);
@@ -367,8 +400,8 @@ const AppButton = new Lang.Class({
 
         this._contextMenuManager = new PopupMenu.PopupMenuManager(this);
         this._appContextMenu = new AppContextMenu(this.actor, this.app);
+        this._appContextMenu.connect('open-state-changed', _onMenuStateChanged);
         this._appContextMenu.actor.hide();
-        this._contextMenuManager.addMenu(this._appContextMenu);
         Main.uiGroup.add_actor(this._appContextMenu.actor);
 
         this._textureCache = St.TextureCache.get_default();
@@ -426,7 +459,7 @@ const AppButton = new Lang.Class({
     _getWindowList: function() {
         let workspace = global.screen.get_active_workspace();
         return this.app.get_windows().filter(function(win) {
-            return win.located_on_workspace(workspace);
+            return !win.skip_taskbar && win.located_on_workspace(workspace);
         });
     },
 
@@ -441,24 +474,39 @@ const AppButton = new Lang.Class({
                 this._windowTitle = new WindowTitle(this.metaWindow);
                 this._singleWindowTitle.child = this._windowTitle.actor;
                 this._windowContextMenu = new WindowContextMenu(this.actor, this.metaWindow);
+                this._windowContextMenu.connect('open-state-changed',
+                                                _onMenuStateChanged);
                 Main.uiGroup.add_actor(this._windowContextMenu.actor);
                 this._windowContextMenu.actor.hide();
                 this._contextMenuManager.addMenu(this._windowContextMenu);
             }
+            this._contextMenuManager.removeMenu(this._appContextMenu);
             this._contextMenu = this._windowContextMenu;
         } else {
             if (this._windowTitle) {
                 this.metaWindow = null;
                 this._singleWindowTitle.child = null;
                 this._windowTitle = null;
-                this._windowContextMenu.actor.destroy();
+                this._windowContextMenu.destroy();
                 this._windowContextMenu = null;
             }
             this._contextMenu = this._appContextMenu;
+            this._contextMenuManager.addMenu(this._appContextMenu);
         }
 
     },
 
+    get active() {
+        return this.actor.has_style_class_name('focused');
+    },
+
+    activate: function() {
+        if (this.active)
+            return;
+
+        this._onClicked(this.actor, 1);
+    },
+
     _onClicked: function(actor, button) {
         let menuWasOpen = this._menu.isOpen;
         if (menuWasOpen)
@@ -512,7 +560,7 @@ const AppButton = new Lang.Class({
         global.window_manager.disconnect(this._switchWorkspaceId);
         this._windowTracker.disconnect(this._notifyFocusId);
         this.app.disconnect(this._windowsChangedId);
-        this._menu.actor.destroy();
+        this._menu.destroy();
     }
 });
 
@@ -698,6 +746,7 @@ const WindowList = new Lang.Class({
 
         let layout = new Clutter.BoxLayout({ homogeneous: true });
         this._windowList = new St.Widget({ style_class: 'window-list',
+                                           reactive: true,
                                            layout_manager: layout,
                                            x_align: Clutter.ActorAlign.START,
                                            x_expand: true,
@@ -724,6 +773,7 @@ const WindowList = new Lang.Class({
                                    Lang.bind(this, this._populateWindowList));
                 }
             }));
+        this._windowList.connect('scroll-event', Lang.bind(this, this._onScrollEvent));
 
 	let indicatorsBox = new St.BoxLayout({ x_align: Clutter.ActorAlign.END });
 	box.add(indicatorsBox);
@@ -783,6 +833,7 @@ const WindowList = new Lang.Class({
 
         this._fullscreenChangedId =
             global.screen.connect('in-fullscreen-changed', Lang.bind(this, function() {
+                this._updateKeyboardAnchor();
                 this._updateMessageTrayAnchor();
             }));
 
@@ -806,6 +857,31 @@ const WindowList = new Lang.Class({
         this._groupingModeChanged();
     },
 
+    _onScrollEvent: function(actor, event) {
+        let direction = event.get_scroll_direction();
+        let diff = 0;
+        if (direction == Clutter.ScrollDirection.DOWN)
+            diff = 1;
+        else if (direction == Clutter.ScrollDirection.UP)
+            diff = -1;
+        else
+            return;
+
+        let children = this._windowList.get_children().map(function(actor) {
+            return actor._delegate;
+        });
+        let active = 0;
+        for (let i = 0; i < children.length; i++) {
+            if (children[i].active) {
+                active = i;
+                break;
+            }
+        }
+
+        active = Math.max(0, Math.min(active + diff, children.length-1));
+        children[active].activate();
+    },
+
     _groupingModeChanged: function() {
         this._groupingMode = this._settings.get_enum('grouping-mode');
         this._grouped = this._groupingMode == GroupingMode.ALWAYS;
